@@ -1,4 +1,3 @@
-from __future__ import print_function
 import os
 import copy
 import numpy as np
@@ -7,10 +6,10 @@ import keras_eval.metrics as metrics
 import keras_eval.visualizer as visualizer
 
 from math import log
+from keras.utils import generic_utils
 
 
 class Evaluator(object):
-
     OPTIONS = {
         'data_dir': {'type': str, 'default': None},
         'concepts': {'type': list, 'default': None},
@@ -20,24 +19,28 @@ class Evaluator(object):
         'report_dir': {'type': str, 'default': None},
         'combination_mode': {'type': str, 'default': None},
         'id': {'type': str, 'default': None},
+        'concept_dictionary_path': {'type': str, 'default': None},
         'loss_function': {'type': str, 'default': 'categorical_crossentropy'},
         'metrics': {'type': list, 'default': ['accuracy']},
         'batch_size': {'type': int, 'default': 1},
         'verbose': {'type': int, 'default': 0},
+        'data_augmentation': {'type': dict, 'default': None},
     }
 
     def __init__(self, **options):
-        # Be able to load keras_applications models by default
+        # Be able to load Keras_applications models by default
         self.custom_objects = utils.create_default_custom_objects()
 
         for key, option in self.OPTIONS.items():
             if key not in options and 'default' not in option:
-                raise ValueError('missing required option: %s' % (key, ))
+                raise ValueError('missing required option: %s' % (key,))
             value = options.get(key, copy.copy(option.get('default')))
             if key == 'custom_objects':
                 self.update_custom_objects(value)
             elif key == 'combination_mode':
                 self.set_combination_mode(value)
+            elif key == 'concept_dictionary_path' and options.get('concept_dictionary_path') is not None:
+                self.concept_dictionary = utils.read_dictionary(value)
             else:
                 setattr(self, key, value)
             if key == 'id' and options.get('model_path') is not None:
@@ -46,13 +49,15 @@ class Evaluator(object):
 
         extra_options = set(options.keys()) - set(self.OPTIONS.keys())
         if len(extra_options) > 0:
-            raise ValueError('unsupported options given: %s' % (', '.join(extra_options), ))
+            raise ValueError('unsupported options given: %s' % (', '.join(extra_options),))
 
         self.results = None
         self.models = []
         self.model_specs = []
         self.probabilities = None
         self.labels = None
+        self.group_id_dict = {}
+        self.combined_probs = []
 
         if self.model_path is not None:
             self.add_model(model_path=self.model_path)
@@ -67,7 +72,8 @@ class Evaluator(object):
 
     def add_model(self, model_path, specs_path=None, custom_objects=None):
         self.update_custom_objects(custom_objects)
-        model, model_spec = utils.load_model(model_path=model_path, specs_path=specs_path, custom_objects=self.custom_objects)
+        model, model_spec = utils.load_model(model_path=model_path, specs_path=specs_path,
+                                             custom_objects=self.custom_objects)
         self.models.append(model)
         self.model_specs.append(model_spec)
 
@@ -90,6 +96,9 @@ class Evaluator(object):
             raise ValueError('Error: invalid option for `combination_mode` ' + str(mode))
 
     def set_concepts(self, concepts):
+        for concept_dict in concepts:
+            if 'label' not in concept_dict.keys() and 'id' not in concept_dict.keys():
+                raise ValueError('Incorrect format for concepts list. It must contain the fields `id` and `label`')
         self.concepts = concepts
 
     def _get_complete_image_paths(self, filenames):
@@ -98,7 +107,7 @@ class Evaluator(object):
             image_paths.append(os.path.join(self.data_dir, filename))
         return image_paths
 
-    def evaluate(self, data_dir=None, top_k=1, filter_indices=None, confusion_matrix=False,
+    def evaluate(self, data_dir=None, top_k=1, filter_indices=None, confusion_matrix=False, data_augmentation=None,
                  save_confusion_matrix_path=None):
         '''
         Evaluate a set of images. Each sub-folder under 'data_dir/' will be considered as a different class.
@@ -109,39 +118,81 @@ class Evaluator(object):
             top_k: The top-k predictions to consider. E.g. top_k = 5 is top-5 preds
             filter_indices: If given take only the predictions corresponding to that indices to compute metrics
             confusion_matrix: True/False whether to show the confusion matrix
+            It includes the addition of data_augmentation as an argument. It is a dictionary consisting of 3 elements:
+            - 'scale_sizes': 'default' (4 scales similar to Going Deeper with Convolutions work) or a list of sizes.
+            Each scaled image then will be cropped into three square parts.
+            - 'transforms': list of transforms to apply to these crops in addition to not
+            applying any transform ('horizontal_flip', 'vertical_flip', 'rotate_90', 'rotate_180', 'rotate_270' are
+            supported now).
+            - 'crop_original': 'center_crop' mode allows to center crop the original image prior do the rest of
+            transforms, scalings + croppings.
+
             save_confusion_matrix_path: If path specified save confusion matrix there
-            combination_mode: Ways of combining the model's probabilities to obtain the final prediction.
-                'maximum': predictions are obtained by choosing the maximum probabity from each class
-                'geometric': predictions are obtained by a geometric mean of all the probabilities
-                'arithmetic': predictions are obtained by a arithmetic mean of all the probabilities
-                'harmonic': predictions are obtained by a harmonic mean of all the probabilities
+
 
         Returns: Probabilities computed and ground truth labels associated.
 
         '''
-        self.data_dir = data_dir or self.data_dir
 
+        self.data_dir = data_dir or self.data_dir
+        self.data_augmentation = data_augmentation or self.data_augmentation
         if self.data_dir is None:
             raise ValueError('No data directory found, please specify a valid data directory under variable `data_dir`')
         else:
-            # Create Keras image generator and obtain predictions
-            self.probabilities, self.labels = self._compute_probabilities_generator(data_dir=self.data_dir)
-
             # Create dictionary containing class names
             if self.concepts is None:
-                self.concepts = utils.create_concepts_default(self.num_classes)
+                self.concepts = utils.get_default_concepts(self.data_dir)
 
             # Obtain labels to show on the metrics results
             self.concept_labels = utils.get_concept_items(self.concepts, key='label')
 
+            if hasattr(self, 'concept_dictionary'):
+                if utils.compare_group_test_concepts(self.concept_labels,
+                                                     self.concept_dictionary) and utils.check_concept_unique(self.concept_dictionary):
+                    # Create Keras image generator and obtain probabilities
+                    self.probabilities, self.labels = self._compute_probabilities_generator(
+                        data_dir=self.data_dir, data_augmentation=self.data_augmentation)
+                    self.compute_inference_probabilities(self.concept_dictionary, self.probabilities)
+
+            else:
+                # Create Keras image generator and obtain probabilities
+                self.probabilities, self.labels = self._compute_probabilities_generator(
+                    data_dir=self.data_dir, data_augmentation=self.data_augmentation)
+
             # Compute metrics
             self.results = self.get_metrics(probabilities=self.probabilities, labels=self.labels,
-                                            concept_labels=self.concept_labels, top_k=top_k, filter_indices=filter_indices,
+                                            concept_labels=self.concept_labels, top_k=top_k,
+                                            filter_indices=filter_indices,
                                             confusion_matrix=confusion_matrix,
                                             save_confusion_matrix_path=save_confusion_matrix_path)
 
         return self.probabilities, self.labels
 
+    def compute_inference_probabilities(self, concept_dictionary, probabilities):
+        '''
+        Combines probabilities based on key "group" in concept_dictionary and saves the values in self.probabilities
+
+        Args:
+            concept_dictionary: It is the dictionary which contains all the granular concepts and the mapping with the groups.
+            probabilities: These are computed granular probabilities
+
+        '''
+
+        for concept in concept_dictionary:
+
+            if concept['group'] in self.group_id_dict.keys():
+                self.group_id_dict[concept['group']].append(concept['class_index'])
+            else:
+                self.group_id_dict[concept['group']] = [concept['class_index']]
+
+        self.combined_probs = [[[0.0, ] * len(self.concept_labels)] * len(probabilities[0])]
+        self.combined_probs = np.array(self.combined_probs)
+        for idx, concept_label in enumerate(self.concept_labels):
+            column_numbers = self.group_id_dict[concept_label]
+            for column_number in column_numbers:
+                self.combined_probs[0][:, idx] += probabilities[0][:, column_number]
+        self.probabilities = self.combined_probs
+
     def plot_confusion_matrix(self, confusion_matrix, concept_labels=None, save_path=None):
         '''
 
@@ -158,24 +209,18 @@ class Evaluator(object):
         visualizer.plot_confusion_matrix(confusion_matrix, concepts=concept_labels, save_path=save_path)
 
     def get_metrics(self, probabilities, labels, top_k=1, concept_labels=None, filter_indices=None,
-                    confusion_matrix=False, save_confusion_matrix_path=None, verbose=0):
+                    confusion_matrix=False, save_confusion_matrix_path=None):
         '''
          Print to screen metrics from experiment given probabilities and labels
 
         Args:
             probabilities: Probabilities from softmax layer
             labels: Ground truth labels
-            combination_mode: Ways of combining the model's probabilities to obtain the final prediction.
-                'maximum': predictions are obtained by choosing the maximum probability from each class
-                'geometric': predictions are obtained by a geometric mean of all the probabilities
-                'arithmetic': predictions are obtained by a arithmetic mean of all the probabilities
-                'harmonic': predictions are obtained by a harmonic mean of all the probabilities
             K: A tuple of the top-k predictions to consider. E.g. K = (1,2,3,4,5) is top-5 preds
             concept_labels: List containing the concept_labels
             filter_indices: If given take only the predictions corresponding to that indices to compute metrics
             confusion_matrix: If True show the confusion matrix
             save_confusion_matrix_path: If path specified save confusion matrix there
-            verbose:
 
         Returns: Dictionary with metrics for each concept
 
@@ -200,7 +245,7 @@ class Evaluator(object):
 
         return results
 
-    def _compute_probabilities_generator(self, data_dir=None):
+    def _compute_probabilities_generator(self, data_dir=None, data_augmentation=None):
         '''
 
         Args:
@@ -215,12 +260,28 @@ class Evaluator(object):
         else:
             for i, model in enumerate(self.models):
                 print('Making predictions from model ', str(i))
-                generator, labels = utils.create_image_generator(data_dir, self.batch_size, self.model_specs[i])
-                # N_batches + 1 to gather all the images + collect without repetition [0:n_samples]
-                probabilities.append(model.predict_generator(generator=generator,
-                                                             steps=(generator.samples // self.batch_size) + 1,
-                                                             workers=1,
-                                                             verbose=1)[0:generator.samples])
+                if data_augmentation is None:
+                    generator, labels = utils.create_image_generator(data_dir, self.batch_size, self.model_specs[i])
+                    # N_batches + 1 to gather all the images + collect without repetition [0:n_samples]
+                    probabilities.append(model.predict_generator(generator=generator,
+                                                                 steps=(generator.samples // self.batch_size) + 1,
+                                                                 workers=1,
+                                                                 verbose=1)[0:generator.samples])
+                else:
+                    generator, labels = utils.create_image_generator(data_dir, self.batch_size, self.model_specs[i],
+                                                                     data_augmentation=data_augmentation)
+                    print('Averaging probabilities of %i different outputs at sizes: %s with transforms: %s'
+                          % (generator.n_crops, generator.scale_sizes, generator.transforms))
+                    steps = generator.samples
+                    probabilities_model = []
+                    for k, batch in enumerate(generator):
+                        if k == steps:
+                            break
+                        progbar = generic_utils.Progbar(steps)
+                        progbar.add(k + 1)
+                        probs = model.predict(batch[0][0], batch_size=self.batch_size)
+                        probabilities_model.append(np.mean(probs, axis=0))
+                    probabilities.append(probabilities_model)
 
             self.generator = generator
             self.num_classes = generator.num_classes
@@ -322,14 +383,16 @@ class Evaluator(object):
         # Get Error Indices, Number of Correct Predictions, Number of Error Predictions per Threshold
         if type == 'probability':
             threshold = threshold or np.arange(0, 1.01, 0.01)
-            correct_ind, errors_ind, correct, errors = metrics.get_top1_probability_stats(self.combined_probabilities, labels,
+            correct_ind, errors_ind, correct, errors = metrics.get_top1_probability_stats(self.combined_probabilities,
+                                                                                          labels,
                                                                                           threshold, verbose=0)
             n_total_errors = errors[0]
             n_total_correct = correct[0]
 
         elif type == 'entropy':
             threshold = threshold or np.arange(0, log(probabilities.shape[1], 2) + 0.01, 0.01)
-            correct_ind, errors_ind, correct, errors = metrics.get_top1_entropy_stats(self.combined_probabilities, labels,
+            correct_ind, errors_ind, correct, errors = metrics.get_top1_entropy_stats(self.combined_probabilities,
+                                                                                      labels,
                                                                                       threshold, verbose=0)
             n_total_errors = errors[-1]
             n_total_correct = correct[-1]
@@ -343,7 +406,7 @@ class Evaluator(object):
 
     def get_image_paths_by_prediction(self, probabilities, labels, concept_labels=None, image_paths=None):
         '''
-
+        Return the list of images given its predictions.
         Args:
             probabilities: Probabilities given by the model [n_samples,n_classes]
             labels: Ground truth labels (categorical)
@@ -356,7 +419,6 @@ class Evaluator(object):
                 'harmonic': predictions are obtained by a harmonic mean of all the probabilities
 
         Returns: A dictionary containing a list of images per confusion matrix square (relation ClassA_ClassB)
-
         '''
         self.combined_probabilities = utils.combine_probabilities(probabilities, self.combination_mode)
 
@@ -386,23 +448,28 @@ class Evaluator(object):
 
         return dict_image_paths_concept
 
-    def plot_images(self, image_paths, n_imgs=None, title='', save_name=None):
+    @staticmethod
+    def plot_images(image_paths, n_images=None, title='', n_cols=5, image_res=(20, 20), save_name=None):
         # Works better defining a number of images between 5 and 30 at a time
         '''
 
         Args:
             image_paths: List with image_paths
-            n_imgs: Number of images to show
+            n_images: Number of images to show
             title: Title for the plot
+            n_cols: Number of columns to split the data
+            image_res: Plot image resolution
+            save_name: If specified, will save the plot in save_name path
 
         Returns: Plots images in the screen
 
         '''
-        image_paths = np.array(image_paths)
-        if n_imgs is None:
-            n_imgs = image_paths.shape[0]
 
-        visualizer.plot_images(image_paths, n_imgs, title, save_name)
+        image_paths = np.array(image_paths)
+        if n_images is None:
+            n_images = image_paths.shape[0]
+
+        visualizer.plot_images(image_paths, n_images, title, n_cols, image_res, save_name)
 
     def compute_confidence_prediction_distribution(self, verbose=1):
         '''
@@ -419,7 +486,7 @@ class Evaluator(object):
 
         '''
         if self.probabilities is None:
-            raise ValueError('probabilities value is None, please run a evaluation first')
+            raise ValueError('probabilities value is None, please run an evaluation first')
         return metrics.compute_confidence_prediction_distribution(self.probabilities, self.combination_mode, verbose)
 
     def compute_uncertainty_distribution(self, verbose=1):
@@ -437,24 +504,58 @@ class Evaluator(object):
 
         '''
         if self.probabilities is None:
-            raise ValueError('probabilities value is None, please run a evaluation first')
+            raise ValueError('probabilities value is None, please run an evaluation first')
         return metrics.uncertainty_distribution(self.probabilities, self.combination_mode, verbose)
 
     def plot_top_k_sensitivity_by_concept(self):
         if self.results is None:
-            raise ValueError('results parameter is None, please run a evaluation first')
+            raise ValueError('results parameter is None, please run an evaluation first')
         concepts = utils.get_concept_items(self.concepts, key='label')
         metrics = [item['metrics']['sensitivity'] for item in self.results['individual']]
         visualizer.plot_concept_metrics(concepts, metrics, 'Top-k', 'Sensitivity')
 
     def plot_top_k_accuracy(self):
         if self.results is None:
-            raise ValueError('results parameter is None, please run a evaluation first')
+            raise ValueError('results parameter is None, please run an evaluation first')
         metrics = self.results['average']['accuracy']
         visualizer.plot_concept_metrics(['all'], [metrics], 'Top-k', 'Accuracy')
 
-    def show_results(self, mode='average', csv_path=None, round_decimals=3):
-        if self.results is None:
-            raise ValueError('results parameter is None, please run a evaluation first')
+    def show_results(self, mode='average', round_decimals=3, show_id=True):
+        '''
+        Args:
+            mode: Mode of results. "average" will show the average metrics while "individual" will show metrics by class
+            round_decimals: Decimal position to round the numbers.
+            show_id: Show id in the first column.
 
-        return utils.show_results(self.results, self.concepts, self.id, mode, csv_path, round_decimals)
+        Returns: Pandas dataframe with results.
+
+        '''
+        if self.results is None:
+            raise ValueError('results parameter is None, please run an evaluation first')
+
+        return utils.results_to_dataframe(self.results, self.id, mode, round_decimals, show_id)
+
+    def save_results(self, id, csv_path, mode='average', round_decimals=3, show_id=True):
+        '''
+
+        Args:
+            id: Name of the results evaluation
+            csv_path: If specified, results will be saved on that location
+            mode: Mode of results. "average" will show the average metrics while "individual" will show metrics by class
+            round_decimals: Decimal position to round the numbers.
+            show_id: Show id in the first column.
+
+        Returns: Nothing. Saves Pandas dataframe on csv_path specified.
+
+        '''
+        if self.results is None:
+            raise ValueError('results parameter is None, please run an evaluation first')
+
+        return utils.save_results(self.results, id, csv_path, mode, round_decimals, show_id)
+
+    def ensemble_models(self, input_shape, combination_mode='average', ensemble_name='ensemble', model_filename=None):
+        ensemble = utils.ensemble_models(self.models, input_shape=input_shape, combination_mode=combination_mode,
+                                         ensemble_name=ensemble_name)
+        if model_filename is not None:
+            ensemble.save(model_filename)
+        return ensemble
